"""
Different usefull functions to interact with the database.yml file
"""
import os
import pkg_resources  # part of setuptools
from datetime import datetime
import time
from pathlib import Path

import yaml
from prettytable import PrettyTable
import numpy as np

DATABASE_ACCESSED = False

class Database:
    """Abstraction to interact with the database"""

    def __init__(self, db_path=None):
        if db_path is None:
            db_path = './database.yml'
        self.db_path = db_path
        self._initialise_db()

    @property
    def _database(self):
        """
        Get the informations from the database.yml file.*
        """

        with open(self.db_path) as fin:
            database = yaml.load(fin, Loader=yaml.FullLoader)
        return database

    @property
    def latest_chain_name(self):
        """Get the last chain name"""
        try:
            out = sorted(self._database, key=self.get_datetime)[-1]
        except IndexError:
            print("There is no chain in the Database 'database.yml'")
        return out

    @property
    def count(self):
        """
        *Get the loop number of a chain*
        """
        return int(len(self._database[self.latest_chain_name]))

    def _initialise_db(self):
        """
        *Create a Database if doesn't exist.*
        """
        if not Path(self.db_path).is_file():
            database = {}
            with open(self.db_path, 'w') as fout:
                yaml.dump(database, fout, sort_keys=False)

    def get_datetime(self, chain_name):
        """
        *Get the datetime of the first loop of a chain*
        """
        database = self._database
        time = database[chain_name][0]['datetime']
        datetime_obj = datetime.strptime(time, "%Y-%m-%d %H:%M:%S")

        return datetime_obj

    def initialise_new_chain(self, chain_name):
        """
        *Create a new chain dict{ } in the DB*
        """
        version = str(pkg_resources.require("lemmings_hpc")[0].version)
        database = self._database

        database[chain_name] = [{'lemmings_version': version,
                                 'datetime': str(datetime.now().strftime("%Y-%m-%d %H:%M:%S")),
                                 'safe_stop': False,
                                 'submit_path': './'}]
        with open(self.db_path, 'w') as fout:
            yaml.dump(database, fout, sort_keys=False)

    def initialise_new_loop(self):
        """Create a new loop in a chain"""
        database = self._database
        database[self.latest_chain_name].append({'datetime': str(datetime.now().strftime("%Y-%m-%d %H:%M:%S")),  # pylint: disable=line-too-long
                                                 'safe_stop': False,
                                                 'submit_path': './'})

        with open(self.db_path, 'w') as fout:
            yaml.dump(database, fout, sort_keys=False)

    def add_nested_level_to_loop(self, key):
        """ Add a nested level in database
            TODO:
                1) add loop options
                2) use Nob if more nested levels will be present
        """
        database = self._database
        loop_num = self.get_current_loop_val('loop_count')
        database[self.latest_chain_name][loop_num -1][key] = []

        with open(self.db_path, 'w') as fout:
            yaml.dump(database, fout, sort_keys=False)

    def update_nested_level_to_loop(self, nested_key, key, value):
        """ Add a nested level in database
            TODO: 
                1) add loop options
                2) use Nob if more nested levels will be present
        """
        database = self._database
        loop_num = self.get_current_loop_val('loop_count')
        if database[self.latest_chain_name][loop_num -1][nested_key] == []:
            database[self.latest_chain_name][loop_num -1][nested_key].append({key:value})
        else:
            database[self.latest_chain_name][loop_num -1][nested_key][0][key] = value

        with open(self.db_path, 'w') as fout:
            yaml.dump(database, fout, sort_keys=False)

    def update_loop(self, key, value, index):
        """
        *Update the database of the current folder

        :param index: The loop number of the desired job
        :type index: int
        :param key: The name of the parameter to update or create
        :type key: str
        :param value: The value of the parameter
        :type value: all
        """
        database = self._database
        database[self.latest_chain_name][index - 1][key] = value

        with open(self.db_path, 'w') as fout:
            yaml.dump(database, fout, sort_keys=False)

    def update_current_loop(self, key, value):
        """
        *Update the database of the current folder*

        :param key: The name of the parameter to update or create
        :type key: str
        :param value: The value of the parameter
        :type value: all
        """
        self.update_loop(key, value, self.count)

    def update_previous_loop(self, key, value):
        """
        *Update the database of the current folder*

        :param key: The name of the parameter to update or create
        :type key: str
        :param value: The value of the parameter
        :type value: all
        """
        self.update_loop(key, value, self.count - 1)

    def update_first_loop(self, key, value):
        """
        *Update the database of the current folder*

        :param key: The name of the parameter to update or create
        :type key: str
        :param value: The value of the parameter
        :type value: all
        """
        self.update_loop(key, value, 1)

    def get_loop_val(self, key, index):
        """
        *Get the value of a parameter in a loop of a job.*

        :param key: The name of the parameter to update or create
        :type key: str
        """
        database = self._database
        return database[self.latest_chain_name][index - 1][key]

    def get_current_loop_val(self, key):
        """
        *Get the value of a parameter in a loop of a job.*

        :param key: The name of the parameter to update or create
        :type key: str
        """
        return self.get_loop_val(key, self.count)

    def get_first_loop_val(self, key):
        """
        *Get the value of a parameter in a loop of a job.*

        :param key: The name of the parameter to update or create
        :type key: str
        """
        return self.get_loop_val(key, 1)

    def get_previous_loop_val(self, key):
        """
        *Get the value of a parameter in a loop of a job.*

        :param key: The name of the parameter to update or create
        :type key: str
        """

        return self.get_loop_val(key, self.count - 1)

    def get_chain_names(self):
        """
        *Get chain names from the database*
        """
        return list(self._database.keys())

    def set_progress_quantity(self, value):
        self.update_first_loop('progress_quantity', value)

    def get_current_status(self, keys=None, with_progress = False):
        """
        Function returning the current status of the simulation.
        """
        # keys : is an option that could allow a future control of what to output
        #           but needs to be developed if eventually needed
        status_string = []
        #NOTE: consider using logging package of python

        try:
            database = self._database  # accesses the whole .yaml file
        except FileNotFoundError as excep:
            status_string.append("LemmingsError: " + excep)
            return status_string

        chain_name = self.latest_chain_name
        if chain_name is None:
            raise ValueError("No chain found. Check database.yml file in your current directory ...")

        #--Check first if the database is the main one of a parallel run---#
        parallel_run = False
        try:
            par_dict, = self.get_current_loop_val('parallel_runs')
            parallel_run = True
        except KeyError:
            pass

        if parallel_run:
            if not with_progress:
                table = PrettyTable()
                table.field_names = ["Workflow number", "Status"]
                # match_val = {"start": "S", "hold": "H", "end": 'F'}
                # table.add_rows([[key, match_val[value]] if value in match_val
                #                     else [key, '?'] for key, value in par_dict.items()])

                symbol = {"start": "S", "wait": "W", "end": "F", "error": "E"}
                table.add_rows([(key, symbol.get(val, '?'))
                                for key, val in par_dict.items()])

                # keep for now in case we run with no match issue
                # table.field_names = ["Workflow number", "Status"]
                # for key, value in par_dict.items():
                #     val = "?"
                #     if value == "start":
                #         val = "S"
                #     elif value == "hold":
                #         val = "H"
                #     elif value == "end":
                #         val = "F"
                #     table.add_row([key,val])

                # print("S: Submitted, F: Finished, W: Wait, E: Error")
                # print(table)
                status_string.append("S: Submitted, F: Finished, W: Wait, E: Error")
                status_string.append(table)
                return status_string

            # case with progress variable
            table = PrettyTable()
            table.field_names = ["Workflow number", "Status", 'Progress']
            symbol = {"start": "S", "wait": "W", "end": "F", "error": "E"}
            _string =  [(key, symbol.get(val, '?'), 'NA')
                                for key, val in par_dict.items()]
            main_dir = os.getcwd()
            for ii, keys in enumerate(_string):
                if keys[1] in ['W']:
                    break
                os.chdir(keys[0])
                tmp_db = Database()
                try:
                    progress_var = tmp_db.get_first_loop_val('progress_quantity')
                except KeyError:
                    print("NOTE: progress_quantity not defined")
                    break
                try:
                    tmp_progress = tmp_db.get_current_loop_val(progress_var)
                except KeyError: 
                    # case when progress var not yet added in current loop
                    # we will then instead use the previous loop
                    if tmp_db.get_current_loop_val('loop_count') == 1:
                        break
                    tmp_progress = tmp_db.get_previous_loop_val(progress_var)
                if isinstance(tmp_progress, float):
                    tmp_progress = np.round(tmp_progress,5)
                _string[ii] = (_string[ii][0], _string[ii][1], tmp_progress)
                os.chdir(main_dir)
            table.add_rows(_string)
            status_string.append("S: Submitted, F: Finished, W: Wait, E: Error")
            status_string.append(table)
            return status_string




        def _handle_first_loop(loop, loop_num, keys):
            """ Function rendering the lemmings-hpc status in case of a first loop
                of a chain of a first loop after a --restart
            """
            value_list = []

            for key in keys:
                if key in loop:
                    if key in ['solut_path']:
                        tmp_path = loop[key].split('/')[0]
                        if tmp_path == '.':
                            tmp_path = './'
                        value_list.append(tmp_path)
                    elif key in ['job_id', 'pjob_id']:
                        value_list.append(loop[key])
                    else:
                        value_list.append('Submitted')
                else:
                    value_list.append('Submitted')
            value_list = [str(ii)] + value_list
            
            return value_list
            
        # print()
        # print("Status for chain %s " % (chain_name))
        status_string.append("Status for chain %s " % (chain_name))

        names_keys = keys
        if keys is None:
            # keys = ['solut_path', 'job_id', 'pjob_id', 'dtsum', 'end_cpu_time']
            keys = ['solut_path', 'condition_reached']
            progress_var = 'progress'
            # progress_title = 'progress'
            # progress_scale = 1
            try:
                progress_var = self.get_first_loop_val('progress_quantity')
                # progress_title = self.get_first_loop_val('progress_title')
                # progress_scale = self.get_first_loop_val('progress_scaling_factor')
            except KeyError:
                pass

            keys.append(progress_var)
            keys.extend(['end_cpu_time', 'job_id', 'pjob_id'])

            names_keys = ['Solution path', 'Job end status',
                          'progress', 'CPU time (h)',
                          'job ID', 'pjob ID']

            # if not progress_title == 'progress':
            #     names_keys = [key.replace('progress', progress_title) for key in names_keys]

        end_message = None

        match_cond_reached_keys = {'True': 'ended, finalized',
                                    'False': 'ended, continue',
                                    'None': 'ended, crashed'}
        table = PrettyTable()
        for ii, loop in enumerate(database[chain_name]):
            value_list = []

            # Separate handling of first loop situation
            if self.get_current_loop_val('loop_count') == 1:
                if not 'condition_reached' in loop:
                    # in case condition_reached is present, we went to post_job already
                    # then the handling should be the normal one
                    value_list = _handle_first_loop(loop, ii, keys)
                    table.field_names = ["Loop"] + names_keys
                    table.add_row(value_list)

                    if loop["end_message"] is not None:
                        end_message = customise_end_message(self, ii, loop["end_message"])
                    break

                if not 'end_cpu_time' in loop:
                    # case first loop is running
                    value_list = _handle_first_loop(loop, ii, keys)
                    table.field_names = ["Loop"] + names_keys
                    table.add_row(value_list)
                    break

            if 'restart' in loop:
                if loop['restart']:
                    end_message = None # we reinit the end_message as loop restarted
                    # value_list = ['---']*(len(keys)+1)
                    # table.add_row(value_list)
                    value_list = ['restart'] + ['-----']*len(keys)
                    table.add_row(value_list)
                    value_list = [] # reinit value_list
                    if not 'end_cpu_time' in loop:
                        # for key in keys:
                        #     if key in loop:
                        #         if key in ['solut_path']:
                        #             tmp_path = loop[key].split('/')[0]
                        #             if tmp_path == '.':
                        #                 tmp_path = './'
                        #             value_list.append(tmp_path)
                        #         elif key in ['job_id', 'pjob_id']:
                        #             value_list.append(loop[key])
                        #         else:
                        #             value_list.append('Submitted')
                        #     else:
                        #         value_list.append('Submitted')
                        # value_list = [str(ii)] + value_list
                        value_list = _handle_first_loop(loop, ii, keys)

                        table.field_names = ["Loop"] + names_keys
                        table.add_row(value_list)
                        if loop["end_message"] is not None:
                            end_message = customise_end_message(self, ii, loop["end_message"])
                        break

            if not 'solut_path' in loop and not 'job_id' in loop:
                # JJ: this case won't happen any more I believe
            # print(self.count, ii)
            # if ii == self.count:
                # if 'job_id' in loop:
                #     continue
                if loop["end_message"] is not None:
                    end_message = customise_end_message(self, ii, loop["end_message"])
                break
            else:
                for key in keys:
                    if key in loop:
                        if key in [progress_var]:
                            tmp_key = np.round(loop[key], 4)
                            value_list.append(tmp_key)
                        elif key in ['end_cpu_time']:
                            value_list.append(np.round(loop[key], 3))
                        elif key in ['solut_path']:
                            tmp_path = loop[key].split('/')[0]
                            if tmp_path == '.' and tmp_path is not None:
                                tmp_path = './'
                            value_list.append(tmp_path)
                        elif key in ['condition_reached']:
                            tmp_key = loop[key]
                            value_list.append(match_cond_reached_keys[str(tmp_key)])
                        else:
                            value_list.append(loop[key])
                    else:
                        if 'condition_reached' not in loop: # will only be added in post_job part
                            value_list.append('Submitted')
                        elif key in [progress_var]:
                            value_list.append("NA")
                        elif 'solut_path' not in loop:
                            try:
                                tmp_path = self.get_first_loop_val('solut_path').split('/')[0]
                                if tmp_path == '.' and tmp_path is not None:
                                    tmp_path = './'
                                value_list.append(tmp_path)
                            except KeyError as excep:
                                continue
                        else:
                            value_list.append("NA")
                value_list = [str(ii)] + value_list
                table.field_names = ["Loop"] + names_keys
                table.add_row(value_list)

                # print(loop["end_message"])
                try:
                    if loop["end_message"] is not None:
                        end_message = customise_end_message(self, ii, loop["end_message"])
                except KeyError: # case when post_job still has to run and add end_message to database
                    pass

        status_string.append(table)
        if end_message is not None:
            status_string.append("Lemmings ended: " + end_message)
        
        return status_string

    def safe_access_to_database(self, total_wait = 0, debug = False):
        """ Provide a safe access to performing tasks in a database.
            Idea is to emulate a lock by the combination of access and release
        """
        # Danger is that we end up in endless loop! Need some timeout safety!
        total_wait_max = 100
        if debug:
            total_wait_max = 19
        if total_wait > total_wait_max:
            raise RuntimeError("We waited way too long to access database."
                + "Something is wrong")

        global DATABASE_ACCESSED
        if not DATABASE_ACCESSED:
            DATABASE_ACCESSED = True
            if debug:
                return DATABASE_ACCESSED
        else:
            print("Database is busy: wait 10 s")
            print("Start : %s" % time.ctime())
            time.sleep(10)
            print("End : %s" % time.ctime())
            # raise ValueError("User end")
            self.safe_access_to_database(total_wait= total_wait + 10, debug = debug)


    def release_access_to_database(self):
        """ Function to unlock the access to the database
        """
        global DATABASE_ACCESSED
        DATABASE_ACCESSED = False

def customise_end_message(database, loop_num, end_msg):
    """ Function handling the end message output shown in 'lemmings status'

        Input:
            :database: database class object
            :loop_num: int, number of the loop calling this functionality
        Output:
            :end_message: str, ouput to be provided
    """

    if not isinstance(loop_num, int):
        loop_num = int(loop_num)

    # if not loop_num == 0:
    #     end_message = "\n  Latest loop = %1d \n" % loop_num#(loop_num - 1)
    # else:
    #     end_message = "\n  Latest loop = %1d \n" % loop_num
    end_message = "\n  Latest loop = %1d \n" % loop_num

    try:
        # end_message += ["  Latest job and pjob IDs = "
        #                 + database.get_loop_val('job_id', (loop_num))
        #                 + ' and ' + database.get_loop_val('pjob_id', (loop_num))
        #                 ][0]

        # Starts counting at 1!!
        end_message += ["  Latest job and pjob IDs = "
                        + database.get_loop_val('job_id', (loop_num+1))
                        + ' and ' + database.get_loop_val('pjob_id', (loop_num+1))
                        ][0]
    except KeyError:
        pass

    end_message += '\n  Final status: ' + end_msg
    return end_message


class AbstractLoop:

    # TODO: let's make this abstract

    def __init__(self, condition_reached,
                 job_id=None, pjob_id=None, **kwargs):
        self.condition_reached = condition_reached
        self.job_id = job_id
        self.pjob_id = pjob_id
        self.user_params = kwargs


class Loop(AbstractLoop):

    def __init__(self, solut_path, condition_reached, end_cpu_time,
                 job_id, pjob_id, progress_quantity=None, **kwargs):
        super().__init__(condition_reached, job_id, pjob_id, **kwargs)
        self.solut_path = solut_path
        self.end_cpu_time = end_cpu_time
        self.progress_quantity = progress_quantity


class EndLoop(AbstractLoop):

    def check_sucess(self):
        return self.job_id is not None


class Printer:

    def __init__(self, database):
        self.loops, self.end_loop = self._get_loops(database)

        # create table
        self.table = PrettyTable()
        self.table.field_names = self._get_name_keys()

        # add rows
        for i, loop in enumerate(self.loops):
            info = list(self._get_info(loop))

            self.table.add_row([i] + info)

    def _get_name_keys(self):
        names_keys = ['Loop', 'Solution path', 'Job end status',
                      'progress', 'CPU time (h)',
                      'job ID', 'pjob ID']

        return names_keys

    def _get_loops(self, database):
        chain_name = database.latest_chain_name
        loops = database._database[chain_name]

        loops_ = []
        for i, loop_dict in enumerate(loops):
            if i == 0:
                progress_quantity = loop_dict.get('progress_quantity', None)

            if i == len(loops) - 1:
                Loop_ = EndLoop
            else:
                loop_dict.update({'progress_quantity': progress_quantity})
                Loop_ = Loop

            loops_.append(Loop_(**loop_dict))

        # TODO: check end loop

        return loops_[:-1], loops_[-1]

    def _get_info(self, loop):
        solut_path = self._prettify_solut_path(loop.solut_path)
        condition_reached = self._map_conditioned_reached(loop.condition_reached)
        progress = self._prettify_progress_quantity(loop.user_params.get(loop.progress_quantity, None))
        end_cpu_time = '{:.3f}'.format(loop.end_cpu_time)

        return (solut_path, condition_reached, progress, end_cpu_time,
                loop.job_id, loop.pjob_id)

    def _prettify_solut_path(self, solut_path):
        tmp_path = solut_path.split(os.path.sep)[0]

        # if only `.` makes it prettier
        if tmp_path == '.':
            tmp_path = f'.{os.path.sep}'

        return tmp_path

    def _map_conditioned_reached(self, condition_reached):
        mapping = {True: 'ended, finalized',
                   False: 'ended, continue',
                   None: 'ended, crashed'}

        return mapping[condition_reached]

    def _prettify_progress_quantity(self, progress_value):
        if progress_value is None:
            return 'N/A'

        if type(progress_value) is int:
            return f'{progress_value}'

        # float
        return '{:.4f}'.format(progress_value)

    def print(self):
        print(self.table)



"""
The machine class that contains user parameter
+ machine cmd
+ machine template.
"""
import os
import sys
import subprocess as sbp
from collections import namedtuple
from datetime import datetime, timedelta

import yaml

from lemmings_hpc.chain.path_tool import PathTools
from lemmings_hpc.chain.lemmingjob_base import LemmingsStop

#__all__ = ["Machine"]

def convert(dictionary):
    """
    *Convert a dict( ) to a NamedTuple( ).*
    """
    return namedtuple('GenericDict', dictionary.keys())(**dictionary)

def get_user_params(path):
    """
    Get the user parameters and generate a NamedTuple with it
    """
    with open(path, 'r') as fin:
        tmp = yaml.load(fin, Loader=yaml.FullLoader)
    # check if file is empty (or everything commented)
    if tmp is None:
        raise ValueError("Oops!  The provided YAML file %s is empty" %path)
    return convert(tmp)

def template_parser(user_tuple, queue_template, job_name, pjob_name):
    """
    Machine template parser that remplace some keys by the user parameters
    """
    lines = queue_template["header"].split("\n")
    content = ""
    for line in lines:
        if "-LEMMING-JOB_NAME-" in line:
            line = line.replace("-LEMMING-JOB_NAME-",
                                job_name)
        if "-LEMMING-POSTJOB_NAME-" in line:
            line = line.replace("-LEMMING-POSTJOB_NAME-",
                                pjob_name)
        if "-LEMMING-WALL-TIME-" in line:
            line = line.replace("-LEMMING-WALL-TIME-",
                                str(queue_template["wall_time"]))
        if "-EXEC-" in line:
            line = line.replace("-EXEC-",
                                user_tuple.exec)
        if "-EXEC_PJ-" in line:
            line = line.replace("-EXEC_PJ-",
                                user_tuple.exec_pj)

        content += line + '\n'
    return content


def get_machine_template(user_tuple, path_machine, job_name, pjob_name):
    #pylint: disable=line-too-long
    """
    Get the machine template + cmd and generate a NamedTuple with it
    """
    with open(path_machine) as fin:
        tmp = yaml.load(fin, Loader=yaml.FullLoader)

    # add the batch key, with substitutions
    tmp["queues"][user_tuple.job_queue]["batch"] = template_parser(user_tuple,
                                                                   tmp["queues"][user_tuple.job_queue],
                                                                   job_name,
                                                                   pjob_name)
    tmp["queues"][user_tuple.pjob_queue]["batch"] = template_parser(user_tuple,
                                                                    tmp["queues"][user_tuple.pjob_queue],
                                                                    job_name,
                                                                    pjob_name)

    cmd = convert(tmp["commands"])
    job_template = convert(tmp["queues"][user_tuple.job_queue])
    pj_template = convert(tmp["queues"][user_tuple.pjob_queue])

    try:
        tmp["queues"][user_tuple.visu_queue]["batch"] = template_parser(user_tuple,
                                                                        tmp["queues"][user_tuple.pjob_queue],
                                                                        job_name,
                                                                        pjob_name)
        visu_template = convert(tmp["queues"][user_tuple.visu_queue])
    except:
        visu_template = None

    return cmd, job_template, pj_template, visu_template

class Machine(PathTools):
    """
    Machine class of Lemmings
    """
    def __init__(self,
                 path_yml,
                 chain_name,
                 user_machine = None,
                 base_dir=None):
        """
        :param path_yml: path to the user parameters file {workflow}.yml
        :type path_yml: str
        :param job_name: The name of the current Lemmings job
        :type job_name: str
        :param base_dir: the base directory to initialise the PathTools class.
        :type base_dir: str
        """
        env_var = dict(os.environ)
        PathTools.__init__(self, base_dir)

        self.path_yml = path_yml
        self.job_name = chain_name
        self.pjob_name = chain_name + "_pj"
        

        try:
            if not self.user.job_prefix in ['None','none']:
                self.job_name = self.user.job_prefix + "_" + self.job_name
                self.pjob_name = self.user.job_prefix + "_" + self.pjob_name
        except:
            pass


        if "LEMMINGS_MACHINE" in env_var and user_machine is None:
            self.path_machine = env_var["LEMMINGS_MACHINE"]
        elif user_machine is not None: # should add user machine instead
            self.path_machine = user_machine
        else:
            raise EnvironmentError("There is no 'LEMMINGS_MACHINE' variable in your Environment"
                                   + " (export LEMMINGS_MACHINE=path/...)")

        self.check_correct_machine_inputs()

    @property
    def user(self):
        """
        *Get the usr parameters from {workflow}.yml
        """
        return get_user_params(self.abspath(self.path_yml))

    @property
    def cmd(self):
        """
        *Get the machine commands {machine}.yml
        """
        return get_machine_template(self.user,
                                    self.abspath(self.path_machine),
                                    self.job_name,
                                    self.pjob_name)[0]
    @property
    def job_template(self):
        """
        *Get the job template from {machine}.yml
        """
        return get_machine_template(self.user,
                                    self.abspath(self.path_machine),
                                    self.job_name,
                                    self.pjob_name)[1]
    @property
    def pj_template(self):
        """
        *Get the post-job template from {machine}.yml
        """
        return get_machine_template(self.user,
                                    self.abspath(self.path_machine),
                                    self.job_name,
                                    self.pjob_name)[2]

    @property
    def visu_template(self):
        """
        *Get the post-job template from {machine}.yml
        """
        return get_machine_template(self.user,
                                    self.abspath(self.path_machine),
                                    self.job_name,
                                    self.pjob_name)[3]

    def check_correct_machine_inputs(self):
        """ We will check if the core_nb has been specified in the machine.yml file.
        """
        try:
             core_nb = self.job_template.core_nb
        except AttributeError:
            raise KeyError("Parameter core_nb not found in machine.yml file, please specify it:"
                            + " core_nb = (number of nodes) x (number of processes per node) ."
                        )

    def submit(self,
               batch_name,
               dependency=None,
               submit_path=None):
        """
        Submit a job on a NFS machine.

        :param batch_name: Name of the batch
        :type batch_name: str
        :param dependency: Job ID of the job to be depend with.
        :type dependency: int
        """

        # check if batch exists
        if not os.path.isfile(os.path.join(submit_path,batch_name)):
            raise FileNotFoundError('Batch file not found. Did you activate user_batch in expert_params?')

        if dependency is None:
            out = sbp.run([self.cmd.submit,
                           batch_name],
                          stdout=sbp.PIPE,
                          stderr=sbp.STDOUT,
                          cwd=submit_path)
        else:
            out = sbp.run([self.cmd.submit,
                           self.cmd.dependency + str(dependency),
                           batch_name],
                          stdout=sbp.PIPE,
                          stderr=sbp.STDOUT,
                          cwd=submit_path)

        out = out.stdout
        job_id = out.decode('utf-8')[:-1].split(" ")[-1]
        return job_id

    def cancel(self,
               job_id):
        """
        Cancel a job on a NFS machine.
        """
        sbp.run([self.cmd.cancel,
                 job_id],
                stdout=sbp.PIPE)

    def get_cpu_cost(self, job_id):
        """
        *Get the CPU cost of the previous Run.*

        :param job_id: Job ID of the previous run
        :type job_id: int
        """
        command_cpu = self.cmd.get_cpu_time.replace("-LEMMING-JOBID-", str(job_id))

        core_nb = 1

        try:
             core_nb = self.job_template.core_nb
        except AttributeError:
            pass    
    
        out = sbp.run(command_cpu,
                    shell=True,
                    stdout=sbp.PIPE)
        out = out.stdout
        out = out.decode('utf-8')[:-1].strip().split(" ")[0]
        out_day = 0
        try_sec = False
        if '-' in out: # we have a format with day as D-H:M:S
            out_day = out.split('-')[0]
            out = out.split('-')[1:][0]
        try:
            out = datetime.strptime(out, '%H:%M:%S').time()
        except ValueError:
            try:
                out = datetime.strptime(out, '%M:%S').time()
            except ValueError:
                try_sec = True
                pass
    
        if not try_sec:
            out_sec = (float(out_day) * 24 * 3600
                    +  out.hour * 3600                             #in seconds
                    + out.minute * 60
                    + out.second)
        else:
            try:
                out_sec = float(out)
            except ValueError as excep:
                print("Error, unknown scheduler CPU format. Job stopped due to infinite loop danger. ", excep)
                sys.exit()

        return (core_nb * out_sec)/3600

        # NEED TO HANDLE:
            # 1-00:00:00
            # 00:30:00    hours min seconds
            # 7-01:00:00    day hours min seconds
            # 3:34    min and seconds
            # 3600     seconds



"""
CLI for Lemmings
"""

import os
import click

@click.group()
def cli():
    """
    *Creation of the group that contains all cli commands.*
    """
    pass


@cli.command('run')
@click.argument('workflow', type=str, nargs=1)
@click.option("--status", "-s", type=str, default="start",
              help="Your job status, Expert users only")
@click.option('--yaml', required=False, type=str, default = None,
              help = "Path to .yml file associated with workflow")
@click.option("--parallel", '-p', is_flag=True,
              help="Parallel mode activation, Expert users only")
@click.option("--noverif", is_flag=True)
@click.option("--restart", is_flag=True,
              help="If activated, the last chain will be continued.")
def lemmings_cli(workflow, status, yaml, noverif, restart, parallel):
    """
    This is the command to launch your workflow with Lemmings.
    ' lemmings run {workflow_name} '
    """
    import sys
    import importlib.util
    from lemmings_hpc.cli_helper import (
        get_workflow_py,
        get_workflow_yml,
        custom_name,
        check_cpu_input,
    )

    from lemmings_hpc.chain.database import Database
    from lemmings_hpc.chain.machine import Machine
    from lemmings_hpc.chain.chain import Lemmings
    from lemmings_hpc.chain.lemmingjob_base import LemmingsStop
    from lemmings_hpc.chain.chain import _handle_exception

    #pylint: disable=too-many-function-args

    #get path of the WorkFlow desired
    if workflow.lower().endswith(('.yml', '.py')):
        raise ValueError("A workflow name can't contain an extension like '.yml', "
                         + "please enter a valid workflow name. ")
#                         + "You can use the 'lemmings info workflows' "
#                        + "command to see available workflows")

    try:
        wf_py_path = get_workflow_py(workflow)
    except EnvironmentError as error:
        print("EnivironmentError", error)
        return

    try:
        if yaml is not None:
            wf_yml_path = get_workflow_yml(yaml.split('.yml')[0], user_yaml = True)
        else:
            wf_yml_path = get_workflow_yml(workflow)
    except FileNotFoundError as error:
        print("FileNotFoundError: \n", error)
        return

    database = Database()
    

    # if we are at a restart we wish to continue the exisiting chain
    # therefore we will overwrite the status to something else
    # a new loop must also be initiated
    if restart:
        status = "restart"
        database.initialise_new_loop()
        database.update_current_loop('restart',
                                           True)

    if status == "start":
        chain_name = custom_name()
    else:
        chain_name = database.latest_chain_name.split("_")[-1]
        # database.initialise_new_loop()



    try:
        machine = Machine(path_yml=wf_yml_path,
                          chain_name=chain_name)
    # except FileNotFoundError as file_error:
    #     print("\nThe initialisation of Lemmings went wrong:\n", file_error)
    except EnvironmentError as excep:
        print("\nThe initialisation of Lemmings went wrong:\n")
        print("   -->  " + str(excep) + '\n')
        return
        #sys.exit() # look into replacing this with return!
    except KeyError as excep:
        print("\nThe initialisation of Lemmings went wrong:\n", excep)
        return
        #sys.exit()

    # put status back to start so lemmings know what to do in the chain
    if restart:
        status = "start"

    if status == "start":
        try:
            if not restart:
                database.initialise_new_chain(machine.job_name)
                print("Starting chain named " + machine.job_name)
            else:
                print("Continue chain named " + machine.job_name)
            if not noverif:
                check_cpu_input(machine.user.cpu_limit)
        except ValueError as excep:
            print("\nThe initialisation of Lemmings went wrong:\n", excep)
            return
        except AttributeError as excep:
            print("\nThe initialisation of Lemmings went wrong:\n AttributeError: keyword cpu_limit not defined in .yml file\n")
            return



    loop_count = database.count

    if wf_py_path is not None:
        spec = importlib.util.spec_from_file_location("module.name",
                                                      wf_py_path)
        module = importlib.util.module_from_spec(spec)
        spec.loader.exec_module(module)
        lemming_job = module.LemmingJob(workflow,
                                        machine,
                                        loop_count,
                                        status=status,
                                        cpu_limit=machine.user.cpu_limit)


    lemmings = Lemmings(lemming_job)

    if parallel:
        try:
            lemmings.create_replicate_workflows()
        except LemmingsStop as stop:
            _handle_exception(lemmings.lemmings_job,lemmings.database,
                            stop, lemmings_stop= True)
            print(stop)
            sys.exit()


    lemmings.run()

@cli.command('kill')
def lemmings_kill():
    """Kill the current job and pjob of lemmings"""
    import os
    import yaml
    from lemmings_hpc.chain.database import Database
    import subprocess as sbp

    database = Database()
    # chain_name = database.latest_chain_name

    env_var = dict(os.environ)
    try:
        machine_path = env_var["LEMMINGS_MACHINE"]
    except KeyError:
        print("No LEMMINGS_MACHINE environment variable specified")
        return

    with open(machine_path) as fin:
        machine_dict = yaml.load(fin, Loader=yaml.SafeLoader)

    try:
        job_id = database.get_loop_val('job_id', 0) #latest loop
        pjob_id = database.get_loop_val('pjob_id', 0)
    except KeyError:
        print("No job and / or post job to kill in current database loop")
        return

    print("Killing the current job and post-job...")

    try:
        #print(type(machine_dict['commands']['cancel']))
        machine_dict['commands']['cancel']
    except KeyError:
        print("No 'cancel' command specified in the machine file")
        return

    sbp.run([machine_dict['commands']['cancel'],
             job_id],
            stdout=sbp.PIPE)
    print("   job  "+ str(job_id) + " was killed")
    sbp.run([machine_dict['commands']['cancel'],
             pjob_id],
            stdout=sbp.PIPE)
    print("   pjob "+ str(pjob_id) + " was killed")




@cli.command('safestop')
def lemmings_safestop():
    """
    Lemmings finish properly the current loop and stop.
    """
    from lemmings_hpc.chain.database import Database
    
    database = Database()
    database.update_loop('safe_stop', True, 0) #latest


@cli.command('endlog')
# @click.option("--job", "-j", type=str, default=None,
#               help="The job name you want to show")
def lemmings_endlog():
    """
    Show the lemmings log of the actual/most_recent job upon finishing.
    """
    import os
    from lemmings_hpc.chain.database import Database
    db = Database()
    # database = db._database
    chain_name = db.latest_chain_name
    if chain_name is None:
        print("No lemmings database found. Check your current directory...")
    else:
        with open(os.path.join(chain_name, chain_name + ".log"), 'r') as fin:
            print(fin.read())

@cli.command('status')
# @click.option("--timeformat", "-tf", required=False , type=click.Choice(['s','ms']), default= 'ms',
#               help="Simulation time format")
@click.option("--database", "-db", required=False , type=str, default = None,
              help="Path to database  YAML file to read")
@click.option("--progress", '-p', is_flag=True,
              help="If activated, the latest progress will also be shown.")
# @click.option("--job", "-j", type=str, default=None,
#               help="The job name you want to show")
def lemmings_status(database, progress):
    """
    Show the status during runtime
    """
    from lemmings_hpc.chain.database import Database

    try:
        if database is not None:
            if os.path.isdir(database): # try find database.yml in directory
                database = os.path.join(database, 'database.yml')
            if not os.path.isfile(database):
                raise FileNotFoundError(database + " does not exist")
        else:
            if not os.path.isfile("database.yml"):
                raise FileNotFoundError("database.yml does not exist")
    except FileNotFoundError as excep:
        print("Error: ", excep)
        return

    if database is not None:
        db = Database(database)
    else:
        ## this init will generate a database.yml file,
        ## so if not present we should check before
        db = Database()

    try:
        _ = [print(string) for string in db.get_current_status(with_progress = progress)]
    except ValueError as excep:
        print("ValueError:", excep)
        return
    except TypeError as excep:
        print("Database currently not accessible, try again shortly")
        print("Make sure it is not corrupted")
        return # might be that database chain not found instead!!!
    except KeyError as excep:
        print("Database currently not accessible, try again shortly")
        return


# @cli.command('get_tuto_files')
# def lemmings_get_tuto():
#     """
#     Recover tuto lemmings/barbatruc files
#     """
#     import os
#     import shutil
#     from pathlib import Path
#     from pkg_resources import resource_filename
#     tuto_path = Path(resource_filename('lemmings', "")) / "example" / "barbatruc"
#     shutil.copytree(tuto_path, os.getcwd() + '/tuto_lemmings')

@cli.command('clean')
@click.option("--database", "-db", required=False , type=str, default = None,
              help="Path to database  YAML file to read")

def lemmings_clean(database):
    """
    Clean lemmings run files in current folder
    """
    import os
    import sys
    import shutil
    from lemmings_hpc.chain.database import Database

    # need to add option to have other database.yml name to look at
    # cfr. lemmings status of FEATURE/parallel branch

    lst_remove = ["__pycache__", "batch_job", "batch_pjob", "database.yml"]

    def remove(my_path):
        """
        Clean removal of file/folders based on the database.yml file info
        """
        if os.path.exists(my_path):
            if os.path.isfile(my_path):
                print("> Remove file: %s" % my_path)
                os.remove(my_path)
                return

            if os.path.isdir(my_path):
                print("> Remove folder: %s" % my_path)
                shutil.rmtree(my_path)
                return

    try:
        if database is not None:
            if not os.path.isfile(database):
                raise FileNotFoundError(database + " does not exist")
            lst_remove = [string.replace('database.yml',database) for string in lst_remove]
        else:
            if not os.path.isfile("database.yml"):
                raise FileNotFoundError("database.yml does not exist")
    except FileNotFoundError as excep:
        print("Error: ", excep)
        return

    if database is not None:
        db = Database(database)
    else:
        ## this init will generate a database.yml file,
        ## so if not present we should check before
        db = Database()

    try:
        db.latest_chain_name
    except UnboundLocalError as excep:
        print("'lemmings clean' aborted")
        return
    except TypeError as excep:
        print("'lemmings clean' aborted")
        return 2

    # Todo: add later option to remove only latest chain name
    try:
        chain_names = db.get_chain_names()
    except TypeError as excep:
        print("ValueError:", excep)
        print("Your file database.yml is corrupted, I can't automatically clean this folder\n")
        sys.exit()

    lst_remove.extend(chain_names)

    # currently only remove WF_xxx folders associated with latest chain
    try:
        par_dict, = db.get_current_loop_val('parallel_runs')
        lst_remove.extend(list(par_dict))
    except KeyError:
        pass

    for path in lst_remove:
        remove(path)


#Main function called by setup.py
main = click.CommandCollection(sources=[cli], help="This is the help of lemmings")



import yaml
from jsonschema import validate
from opentea.noob.validate_light import validate_light

with open("cobalt.yml", 'r') as fin:
    dict_ = yaml.load(fin, Loader=yaml.FullLoader)

with open("machine_schema.yml", 'r') as fin:
    dict_schema = yaml.load(fin, Loader=yaml.FullLoader)


#validate_light(dict_, dict_schema)
validate(dict_, dict_schema)

flagSCHEMA = 0
if flagSCHEMA:
    from genson import SchemaBuilder    
    builder = SchemaBuilder(schema_uri="http://json-schema.org/draft-04/schema#")
    builder.add_object(dict_)
    schema = builder.to_schema()  

    flagWriteYaml = 0
    if flagWriteYaml:
        with open('test_schema.yml', 'w') as yaml_out:
            yaml.dump(schema,yaml_out,indent = 2)





"""Module helper for the CLI of lemmings"""
import os
import random
import string
import glob
from pathlib import Path
from pkg_resources import resource_filename
from lemmings_hpc.chain.path_tool import PathTools

# We have to avoid offensive names, here in french
FORBIDDEN_NAMES = [
    "", "PUTE", "PEDE", "BITE", "CACA", "ZIZI", "CUCU", "PIPI", "CONE",
    "CULE", "PENE", "SODO", "SUCE", "SUCA", "KUKU", "CUNI"]

def custom_name():
    """
    Random Name generator --> Cons + Vowels + Cons + Vowels + 2*Int
    """
    vowels = list("AEIOU")
    consonants = list(set(string.ascii_uppercase) - set(vowels))
    name = ""
    while name in FORBIDDEN_NAMES:
        name = ""
        name += random.choice(consonants)
        name += random.choice(vowels)
        name += random.choice(consonants)
        name += random.choice(vowels)

    name += str(random.randrange(1, 10))
    name += str(random.randrange(1, 10))
    return name


def get_workflow_py(workflow):
    """
    *Get the path of the {workflow}.py file.*

    :param workflow: workflow name
    :type workflow: str
    """
    path_tool = PathTools()
    env_var = dict(os.environ)
    catch = False

    if Path("./" + workflow + ".py").is_file():
        wf_py_path = path_tool.abspath(workflow + ".py")
        catch = True
    if catch is False and "LEMMINGS_WORKFLOWS" in env_var:
        wf_py_path = env_var["LEMMINGS_WORKFLOWS"]
        if Path(wf_py_path).is_file():
            catch = True
        else:
            raise EnvironmentError("\n ENVIRONMENT VAR :\nunable to open file at " + wf_py_path)
    if catch is False:
        wf_py_path = Path(resource_filename('lemmings_hpc', "")) / "chain" / "workflows" /  str(workflow + ".py") #pylint: disable=line-too-long
        if Path(wf_py_path).is_file():
            catch = True
        else:
            raise EnvironmentError("\nFOLDER LEMMING : unable to open file at " + str(wf_py_path))

    if wf_py_path is not None:
        return wf_py_path

    raise EnvironmentError("Your workflow doesn't exist in :\n"
                           + "  - current directory\n"
                           + "  - Environment variable ('LEMMING_WORKFLOWS')\n"
                           + "  - Folder workflows of lemmings\n")


def get_workflow_yml(workflow, user_yaml = False):
    """
    *Get the path of the {workflow}.yml file.*

    :param workflow: workflow name
    :type workflow: str
    """
    path_tool = PathTools()
    if user_yaml:
        if Path(workflow + ".yml").is_file():
            wf_yml_path = path_tool.abspath(workflow + ".yml")
        else:
            raise FileNotFoundError("Oops!  Couldn't find the %s.yml file relative to your current directory. Please generate it." % workflow)
    else:
        if Path("./" + workflow + ".yml").is_file():
            wf_yml_path = path_tool.abspath(workflow + ".yml")
        else:
            raise FileNotFoundError("Oops!  Couldn't find the %s.yml file in your current directory. Please generate it." % workflow)

    return wf_yml_path


def check_cpu_input(cpu_limit):
    """
    *Ask a confirmation of the cpu_limit to the user*

    :param cpu_limit: the cpu limit found in the {workflow}.yml file
    :type cpu_limit: float
    """
    cpu_confirm = input("Your CPU limit is " + str(cpu_limit) + " [hours]."
                        + " Confirm it typing the same value: ")

    if int(cpu_confirm) != int(cpu_limit):
        raise ValueError("Confirmation failed: " + str(cpu_confirm) + " != " + str(cpu_limit))


def find_all_wf():
    """
    *Find all available workflows.*
    They can be in the current directory,
    workflows folder of lemmings or in a environment variable.
    """
    path_workflows = {"current": {},
                      "env_var": {},
                      "lemmings": {}}

    files_wf = []
    env_var = dict(os.environ)

    #find in the workflows folder of lemmings
    path_folder_wf = Path(resource_filename('lemmings_hpc', "")) / "chain" / "workflows"
    for file in glob.glob(str(path_folder_wf) + "/*.py"):
        if "__init__" not in file:
            files_wf.append(file.split("/")[-1])
            path_workflows["lemmings"][files_wf[-1].split('.py')[0]] = file

    #find in current directory 
    for file in glob.glob("*.py"):
        if file in files_wf: # -> requires to be def
            path_workflows["current"][file.split('/')[-1].split('.py')[0]] = file

    #find in environment variable
    if "LEMMINGS_WORKFLOWS" in env_var:
        for var in env_var["LEMMINGS_"]:
            if var.split('/')[-1] in files_wf:
                path_workflows["env_var"][var.split('/')[-1].split('.py')[0]] = var

    return path_workflows



#!/usr/bin/env python
# pylint: disable=no-value-for-parameter
# -*- coding: utf-8 -*-
"""path_tools.py

Concentrate path and file manipulaiton tools.  Any
OpenTEA class needing these tools should inherit it.

Created Nov 2016 by COOP team
"""

from __future__ import absolute_import

import os
import shutil
import logging
import errno
import glob

from os.path import (abspath, basename, isabs, isfile, normpath,
                     realpath, isdir, join, relpath, samefile)

__all__ = ["PathTools"]


class PathTools:
    """Concentrate path manipulation tools"""
    def __init__(self, base_dir=None, **kwargs):
        self.base_dir = os.getcwd() if base_dir is None else base_dir
        self.rel_dir = self.base_dir
        self.log = logging.getLogger(__name__)
        self.log.debug("Instance of PathTools created")
        self.log.debug("    base_dir: " + self.base_dir)
        self.log.debug("    rel_dir : " + self.rel_dir)
        self.__dict__.update(**kwargs)

    def abspath(self, *args):
        """Get path as clean absolute

        If args starts with absolute path, just abspath it.  Else, consider
        from self.base_dir, then abspath.  Note: os.path.abspath also performs
        normpath, which "cleans" any ../ and such.
        """
        interpret_path_from = (join(*args) if isabs(args[0]) else
                               join(self.base_dir, *args))
        return normpath(realpath(abspath(interpret_path_from)))

    def relpath(self, *args):
        """Get path as clean relative from base_dir"""
        return './' + normpath(relpath(realpath(self.abspath(*args)),
                                       self.rel_dir))

    def _file_checks(self, src, dest):
        """Generic checks before moving or copying src to dest

        - If dest is a folder, src file name is kept. Otherwise, given file
          name is used.
        - If source and dest are the same file, no error is raised but a
          warning is printed.

        Return go_code (True: continue. False: skip) and destination
        """
        dest = self.abspath(dest)
        if isdir(dest):
            dest = join(dest, basename(src))

        if isfile(src) and isfile(dest) and samefile(src, dest):
            # shutil.copy doesn't like same files
            self.log.warning(
                "No copy for the same file: {0} and {1}".format(src, dest))
            return False, dest

        return True, dest

    def copy_file(self, src, dest):
        """Copy file src to dest"""
        copy_is_ok, dest = self._file_checks(src, dest)
        if copy_is_ok:
            self.log.debug("Copying {0} to {1}".format(src, dest))
            shutil.copy(src, dest)
        return dest

    def move_file(self, src, dest):
        """Move file src to dest"""
        move_is_ok, dest = self._file_checks(src, dest)
        if move_is_ok:
            self.log.debug("Moving {0} to {1}".format(src, dest))
            if isfile(dest):
                os.remove(dest)
            shutil.move(src, dest)
        return dest

    def copy_dir(self, src, dest):
        """Copy dir src to dest
        - If dest ends with /, src dir is copied in dest dir.  Otherwise, src
          dir is copied *as* dest dir.  This mirrors the `cp` command behavior.
        - If dest is a regular file, raise an error.
        - If source and dest are the same dir, no error is raised but a warning
          is printed.
        """
        if dest[-1] == '/':
            dest = join(dest, basename(src))

        dest = self.abspath(dest)
        if isfile(dest):
            self.log.error("Trying to create/populate directory {0}, but it "
                           "is already a regular file".format(dest))
            raise OSError(errno.EEXIST)

        # shutil.copy doesn't like same files
        if isdir(src) and isdir(dest) and samefile(src, dest):
            self.log.warning("No copy for the same file: {0} and {1}".
                             format(src, dest))
            return

        if not isdir(dest):
            os.mkdir(dest)

        self.log.debug("Copying {0} to {1}".format(src, dest))
        for path in os.listdir(src):
            if isdir(path):
                self.copy_dir(join(src, path), join(dest, path))
            else:
                self.copy_file(join(src, path), dest)

    def ensure_dir(self, directory):
        """Ensure that directory exists"""
        new_dir = self.abspath(directory)
        self.log.debug("Ensuring existence of " + new_dir)
        if not isdir(new_dir):
            os.mkdir(new_dir)
        return new_dir

    def get_file_list(self, path, filter_search="", output="listonly"):
        """ return the list of files in a path (compulsory)
        matching a filter string (facultative).
        output (facultative) can be : - listonly : a list of the files
                                      - relative : a list relative pathes
                                      - absolute : a list of absolute path
        """
        lresult = []
        glob_arg = os.path.join(path, filter_search)
        list_dir = glob.glob(glob_arg)

        if output == "relative":
            for file_name in list_dir:
                lresult.append(os.path.relpath(file_name))
        if output == "listonly":
            for file_name in list_dir:
                lresult.append(os.path.relpath(file_name, path))
        if output == "absolute":
            lresult = list_dir

        lresult = sorted(lresult)
        self.log.debug("Looking for files " + filter_search + "in " + path)
        self.log.debug(".. List of files :" + str(lresult))

        return lresult


""" The lemmings base class containing the methods which get inherited by LemmingJob
"""
import os
import subprocess
import numpy as np
from nob import Nob


from lemmings_hpc.chain.path_tool import PathTools
from lemmings_hpc.chain.database import Database

class LemmingJobBase():

    def __init__(self,
                 workflow,
                 machine,
                 loop_count,
                 status="start",
                 cpu_limit=None,
                 base_dir=None
                 ):

        self.pathtools = PathTools(base_dir)
        self.database = Database()
        self.workflow = workflow
        self.machine = machine
        self.status = status    #start, spawn_job, post_job, exit
        self.cpu_limit = cpu_limit
        self.loop_count = loop_count
        self.end_message = None

        if self.status == 'start':
            self.database.update_current_loop('loop_count',
                                            self.loop_count)
            self.database.update_current_loop('end_message',
                                            self.end_message)

        # can be overruled in expert  params but not for now
        #self.max_parallel_workflows = 2

    """
    A lemming job follows always the same pattern.
    START > SPAWN JOB> POST JOB > SPAWN JOB > POST JOB > EXIT

    each step can be customized in the present class.
    e.g. you control the nb. of SPAWN JOB>POST JOB with the 'Check on end` function.


                 Prior to job  +---------+             Prepare run
                     +--------->SPAWN JOB+---------------------+
                     |         +------^--+                     |
                     |                |                      +-v------+
                   True               |                      |POST JOB|
    +-----+          |                |                      +--------+
    |START+--->Check on start         |                          v
    +-----+          |                +---------------False-Check on end
                   False            Prior to new iteration       +
                     |                                         True
                     |                                           |
                     |                                           |
                     |           +----+                          |
                     +---------->|EXIT|<-------------------------+
               Abort on start    +----+                After end job

    you can use the database if you need to store info from one job to the other.

    The following definition of methods allows a single lemmings run to be performed without any other user input
    except for the required .yml file information.

    """
    def prior_to_job(self):
        """
        Function that prepares the run when the user launches the Lemmings command.
        """

        pass

    def abort_on_start(self):
        """
        What lemmings does if the criterion is reached in the first loop.
        """

        pass

    def prepare_run(self):
        """
        Prepare the run before submission.
        """

        pass

    def prior_to_new_iteration(self):
        """
        Prepare the new loop specific actions if criterion is not reached.
        """

        pass

    def after_end_job(self):
        """
        Actions just before lemmings ends.
        """

        pass


    def check_on_start(self):
        """
        Verify if the condition is already satisfied before launching a lemmings chain.

        Function returns a boolean which starts the chain run. Default set to True.

        A minimum required action is to set the 'start_cpu_time' so that lemmings can check
        if the max cpu condition is reached.

        """

        self.database.update_current_loop('start_cpu_time', 0.)
        start_chain = True

        return start_chain


    def check_on_end(self):
        """
        Verifications after each job loop

         The function check_on_end needs to return a boolean (default True) with three options:
             - False: we continue lemmings
             - True: target reached, we stop lemmings (default setting)
             - None: crash, we stop lemmings

        Default verification by lemmings:
             - is the cpu condition (.yml file) reached?
         """

        condition_reached = True

        return condition_reached

    def monitor_replicate_workflows(self, debug = False):
        """ Function that monitors the number of parallel jobs currently run

            Idea is to update the job status in the main database.yml.
                Status should be either: start, hold, end
                So:
                    1) I must be able to access the main database.yml
                    2) I must be able to submit new lemming jobs
                    3) I must update database status
        """

        wf_dir = os.getcwd() # need ton know where I am to update the main database
        wf_current = wf_dir.split('/')[-1]
        # print("Current workflow", wf_current)

        # absolute path to main directory --> considers a fixed structure!
        base_run_path = "/" + os.path.join(*self.machine.path_yml.split('/')[0:-1])
        os.chdir(base_run_path)
        # print(os.getcwd())
        # Lock access to database
        try:
            self.database.safe_access_to_database(debug = debug)
        except RuntimeError as excep:
            raise LemmingsStop("Monitoring of lemmings chains stopped", excep)
        # Do stuff with our database
        db_info = Nob(self.database._database)
        # Modify status of current workflow: from start to end
        #db_wf_path, = db_info.find(wf_current)
        #print(db_info[db_wf_path][:])
        self.database.update_nested_level_to_loop('parallel_runs',wf_current, 'end')
        db_info = Nob(self.database._database) # required for full status
        #print(db_info[db_wf_path][:]) # doesn't work directly, need to reread it somehow


        # Find next lemmings chain to launch through 'hold' keyword
        status_list = []
        db_subtree = db_info[self.database.latest_chain_name]
        #print(db_subtree)
        for key in db_subtree.parallel_runs[:][0].keys():
            #print(key)
            #print(db_info[key][:])
            if db_subtree[key][:] == 'wait':
                print("We found new job to launch ", key)
                self.database.update_nested_level_to_loop('parallel_runs',key, 'start')
                os.chdir(key)
                if not debug:
                    subprocess.call("lemmings-hpc run " + self.workflow + " --noverif"
                            + " --yaml=../"+self.workflow+".yml" , shell = True)
                os.chdir(base_run_path) # back to main directory
                if debug:
                    self.database.release_access_to_database()
                    os.chdir(wf_dir)
                    return 'Subprocess should have been launched'
                break
            status_list.append(db_subtree[key][:])
            self.database.update_current_loop("end_message", "All lemmings chains"
                                        +  " have been submitted")
        if len(np.unique(status_list)) == 1:
            self.database.update_current_loop("end_message", "All lemmings chains have ended")

        # Release access to database
        self.database.release_access_to_database()
        os.chdir(wf_dir)

class LemmingsStop(Exception):
    """ Definition of a class to allow exit of Lemmings safely upon exceptions
    """

    pass



#!/usr/bin/env python
from setuptools import setup, find_packages
from os.path import splitext, basename
import os.path as path
from glob import glob

this_directory = path.abspath(path.dirname(__file__))
with open(path.join(this_directory, 'README.md'), encoding='utf-8') as f:
     long_description = f.read()

with open(path.join(this_directory, 'VERSION'), encoding='utf-8') as f:
    version = f.read()

with open('LICENSE.txt') as f:
     license = f.read()

setup(name='lemmings-hpc',
    version=version,
    description='Flexible chaining of jobs on hpc with workflows',
    author='Thibault Gioud, Jimmy-John Hoste',
    author_email="coop@cerfacs.fr",
    url="",
    license ="MIT License",
    license_files = license,
    long_description=long_description,
    long_description_content_type="text/markdown",
    keywords=["hpc","job chaining", "worfklows"],
    classifiers=[
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "License :: OSI Approved :: MIT License",
    ],
    # packages=find_packages(include=['lemmings']),
    install_requires=[
        "click",
        "prettytable>=2.0.0, <3.0",
        "numpy",
        "PyYAML",
      ],
    tests_require=[
        'pytest',
        'pytest-mock'
      ],

    packages=find_packages("src"),
    package_dir={"": "src"},
    py_modules=[splitext(basename(path))[0] for path in glob('src/*.py')],
    include_package_data=True,
    entry_points={
        "console_scripts": ["lemmings-hpc = lemmings_hpc.cli:main"]
        }
     )



"""
This script executes the fundamentals functions of lemmings workflows.
"""
import sys
import os
import traceback
import shutil
import subprocess
#import logging
#import yaml
from prettytable import PrettyTable
from nob import Nob
import numpy as np

from lemmings_hpc.chain.database import Database
from lemmings_hpc.chain.lemmingjob_base import LemmingsStop

class Lemmings():
    """
    The mother class of Lemmings.
    """
    def __init__(self,
                 lemmings_job):
        """
        :param lemmings_job: An Object that contains all actions/
                             made in the different classical lemmings function:
                             --> function that check conditon(s)
                             --> function that do some actions before update status
                            For example, function "start_to_spawn_job()" can save a file, a fig ...
        """
        self.lemmings_job = lemmings_job
        self.database = Database()


    def run(self):
        """*Submit the chain of computations.*"""

        chain_name = self.lemmings_job.machine.job_name
        if not os.path.exists(chain_name):
            os.mkdir(chain_name)

        # logging.basicConfig(filename="lemmings.log", level=logging.INFO, format='%(message)s')

        # if self.lemmings_job.status == "start":
        #     logging.info("~~~~~~ START NEW Lemmings Chain ~~~~~~~\n")
        # else:
        #     logging.info("\n####### START NEW Lemmings loop #######")

        while self.lemmings_job.status != "exit":
            self.next()
        # else:
        #     print("Lemmings stopped: ", self.lemmings_job.end_message)
                # Could do print: Lemmings status = "started, aborted etc"
                # and in init of end_message = "Starting"
                # the other prints seem to be in log file instead ..
                # so do smth in the CLI? perhaps better to do so?


    def next(self):
        """
        Execute all necessary functions depending on its status
        There are 2 kind of function:
            - Functions that check some conditions
            - Functions that pass from a status to another
    ::

                     - - - > spawn_job < - - -
                    |             |             |
                    |             |             |
                  start             - - - - > post_job
            <check_condition>             <check_condition>
                    |                           |
                    |                           |
                    |                           |
                     - - - - - > Exit < - - - -
        """
        if self.lemmings_job.status == "start":
            _perform_lemmings_job_start(self)

        elif self.lemmings_job.status == "spawn_job":
            _perform_lemmings_job_spawn(self)

        elif self.lemmings_job.status == "post_job":
            _perform_lemmings_job_post(self)


    def _check_cpu_cost(self):
        """Check if the CPU limit is reached."""

        #--testing--#
        last_job_id = self.database.get_current_loop_val('job_id')
        last_cpu_time = self.database.get_current_loop_val('start_cpu_time')


        # last_job_id = self.database.get_previous_loop_val('job_id')
        # last_cpu_time = self.database.get_previous_loop_val('start_cpu_time')

        new_cpu_time = self.lemmings_job.machine.get_cpu_cost(last_job_id)
        total_cpu_time = last_cpu_time + new_cpu_time

        # self.database.update_previous_loop('end_cpu_time',
        #                                    total_cpu_time)
        # self.database.update_current_loop('start_cpu_time',
        #                                   total_cpu_time)
       
        self.database.update_current_loop('end_cpu_time',
                                           total_cpu_time)
        #start cpu time should be updated if new loop!
        #--end testing--#

        if total_cpu_time > self.lemmings_job.cpu_limit:
            self.database.update_previous_loop('cpu_reached',
                                               True)
            self.lemmings_job.status = 'exit'
            self.lemmings_job.end_message = "CPU condition reached"

            self.database.update_current_loop('end_message',
                                    self.lemmings_job.end_message)

            return True
        return False


    def _create_batch(self, batch_j="./batch_job", batch_pj="./batch_pjob"):
        """
        Create the batch that will launch the job and postjob loop of lemmings.
        The construction is based on namedtuple that are unique for each machine.
        So the user, if not already done, have to set up those namedtuple for his machine(cluster).
        """

        # The user can take control of this step which can be done through the
        # expert_params object in the workflow's .yml file
        if hasattr(self.lemmings_job.machine.user, 'expert_params'):
            if 'user_batch' in self.lemmings_job.machine.user.expert_params:
                if self.lemmings_job.machine.user.expert_params['user_batch']:
                    return

        batch_job = self.lemmings_job.machine.job_template.batch
        batch_pjob = (self.lemmings_job.machine.pj_template.batch + '\n'
                      + "lemmings-hpc run "
                      + str(self.lemmings_job.workflow)
                      + " -s post_job"
                      + " --yaml=" + self.lemmings_job.machine.path_yml
                      + '\n')

        with open(batch_j, 'w') as fout:
            fout.write(batch_job)
        with open(batch_pj, 'w') as fout:
            fout.write(batch_pjob)


    def write_log_file(self, usefull_keys=None):
        """write the log file"""
        chain_name = self.database.latest_chain_name
        database = self.database._database
        table = PrettyTable()

        if usefull_keys is None:
            usefull_keys = ['datetime', 'job_id', 'pjob_id', 'dtsum', 'end_cpu_time']

        if chain_name is None:
            raise ValueError("No chain found. Check database file in your current directory ...")
        else:
            log_msg = "Lemmings Version : " + str(database[chain_name][0]['lemmings_version']) + '\n\n'

            for i, loop in enumerate(database[chain_name]):
                value_list = []
                for key in usefull_keys:
                    if key in loop:
                        value_list.append(loop[key])
                    else:
                        value_list.append(None)
                value_list = [str(i)] + value_list
                table.field_names = ["Loop"] + usefull_keys
                table.add_row(value_list)
            log_msg += str(table)
            log_msg += "\n\n"

        if database[chain_name][-1]['safe_stop'] is True:
            log_msg += "Lemmings STOP because using 'safe stop' command\n"
        if 'run_crash' in database[chain_name][-2] and database[chain_name][-2]['run_crash'] is True:
            log_msg += "Your run CRASHED, see avbp.o file\n"
        elif 'condition_reached' in database[chain_name][-2] and database[chain_name][-2]['condition_reached']:
            if 'simu_end_time' in database[chain_name][0]:
                log_msg += ("Condition " + str(database[chain_name][0]['simu_end_time'])
                            + " [s] is     REACHED")

        with open(os.path.join(chain_name, chain_name + '.log'), 'w') as fout:
            fout.write(log_msg)


    def create_replicate_workflows(self):
        """ Method that controls the generation of multiple workflows for parallel mode 
            and submits them according to user settings

             Split in smaller functions
                1) check if all is well activated in workflow.yml
                    -> _check_correct_parallel_settings()
                2) perform the workflow copies
                    -> _generate_workflow_replicates
                3) launch workflows
                    3.1) check if max parallel workflow specied
                    3.2) launch workflows
                    3.3) update the database
                    -> _launch_workflows_parallel
                4) raise LemmingsStop as we did what we had to do at this point
        """

        #Check if parameters correctly defined by user in .yml file
        num_workflows = _check_correct_parallel_settings(self.lemmings_job)

        print("Parallel mode enabled")
        self.database.add_nested_level_to_loop('parallel_runs')

        # Generation of different workflow folder replicates
        workflows_list = _generate_workflow_replicates(num_workflows, self.lemmings_job, self.database)
        # Launch workflows from respective folders
        max_par_wf = _launch_workflows_parallel(workflows_list, self.lemmings_job, self.database)


        try:
            raise LemmingsStop("Replicate workflows launched according to,"
                                + " max parallel chains = %3d " % max_par_wf)
        except TypeError:
            raise LemmingsStop("All replicate workflows launched")


def _perform_lemmings_job_start(lemmings):
    """ Function controls the call sequence associated with the 'start' part of lemmings

        Input:
            :lemmings: Lemmings object

        Output:
            :None
    """

    try:
        start_chain = lemmings.lemmings_job.check_on_start()
        if start_chain:
            lemmings.lemmings_job.prior_to_job()
            lemmings._create_batch()
            lemmings.lemmings_job.status = "spawn_job"
        else:
            lemmings.lemmings_job.abort_on_start()
            lemmings.lemmings_job.status = "exit"
    except LemmingsStop as stop:
        _handle_exception(lemmings.lemmings_job, lemmings.database,
                            stop, lemmings_stop = True)
        lemmings.lemmings_job.status = "exit" #-> can probably remove, to check
    except Exception as any_other_exception:
        _handle_exception(lemmings.lemmings_job,lemmings.database,
                            any_other_exception, lemmings_stop = False)


def _perform_lemmings_job_spawn(lemmings):
    """ Function controls the call sequence associated with the 'spawn_job' part of lemmings

        Input:
            :lemmings: Lemmings object

        Output:
            :None
    """

    try:
        # Defined as one of the methods
        lemmings.lemmings_job.prepare_run()
        safe_stop = lemmings.database.get_previous_loop_val('safe_stop')
    except LemmingsStop as stop:
        _handle_exception(lemmings.lemmings_job,lemmings.database,
                            stop, lemmings_stop= True)
        safe_stop = True
    except Exception as any_other_exception:
        _handle_exception(lemmings.lemmings_job,lemmings.database,
                            any_other_exception, lemmings_stop= False)
        safe_stop = True
    
    if safe_stop is False:
        submit_path = lemmings.database.get_current_loop_val('submit_path')
        try:
            job_id = lemmings.lemmings_job.machine.submit(batch_name="batch_job",
                                                    submit_path=submit_path)
        except FileNotFoundError as excep:
            print("LemmingsError:", excep)
            #TODO: pass through _handle_exception in here
            #       and raise LemmingsStop
            sys.exit()
        try:
            pjob_id = lemmings.lemmings_job.machine.submit(batch_name="batch_pjob",
                                                    dependency=job_id,
                                                    submit_path="./")
        except FileNotFoundError as excep:
            print("LemmingsError:", excep)
            #TODO: pass through _handle_exception in here
            #       and raise LemmingsStop
            sys.exit()

        lemmings.database.update_current_loop('job_id',
                                        job_id)
        lemmings.database.update_current_loop('pjob_id',
                                        pjob_id)
    else:
        lemmings.database.update_current_loop('safe_stop',
                                        True)
    lemmings.lemmings_job.status = "exit"


def _perform_lemmings_job_post(lemmings):
    """ Function controls the call sequence associated with the 'post_job' part of lemmings

        Input:
            :lemmings: Lemmings object

        Output:
            :None
    """

    # A lemmings run is finished if
    #       1) the CPU limit is reached
    #       2) the target condition is reached (e.g. simulation end time)
    #       3) the simulation crashed for some reason
    # condition_reached can take 3 values:
    #       - False: we continue lemmings
    #       - True: target reached, we stop lemmings
    #       - None: crash, we stop lemmings


    # 1) check if cpu cost reached
    condition_reached = lemmings._check_cpu_cost()
    # 2) check if target condition reached 3) or crash
    if not condition_reached:
        try:
            condition_reached = lemmings.lemmings_job.check_on_end()

            if condition_reached is True or condition_reached is None:
                #--testing------#
                # lemmings.database.update_previous_loop('condition_reached',
                #                                 condition_reached)
                 #--end testing--#
                lemmings.database.update_current_loop('condition_reached',
                                                condition_reached)                      
                if condition_reached is None:
                    lemmings.lemmings_job.end_message = "Run crashed"
                else:
                    lemmings.lemmings_job.end_message = "Target condition reached"

                lemmings.database.update_current_loop('end_message',
                            lemmings.lemmings_job.end_message)

                lemmings.lemmings_job.after_end_job()
                _check_and_activate_parallel_mode(lemmings.lemmings_job, "monitor")
                lemmings.write_log_file()
                lemmings.lemmings_job.status = "exit"
            else:
                # lemmings.database.update_previous_loop('condition_reached',
                #                             condition_reached)

                #--testing------#
                lemmings.database.update_current_loop('condition_reached',
                                                 condition_reached)
                lemmings.database.initialise_new_loop()
                lemmings.lemmings_job.loop_count += 1 # increment loop count by 1 as we'll start a new spawn job
                lemmings.database.update_current_loop('loop_count',
                                            lemmings.lemmings_job.loop_count)
                lemmings.database.update_current_loop('end_message',
                                                        None)
                lemmings.database.update_current_loop('start_cpu_time',
                                           lemmings.database.get_previous_loop_val('end_cpu_time'))
                #--end testing--#

                lemmings.lemmings_job.prior_to_new_iteration()
                lemmings._create_batch()
                lemmings.lemmings_job.status = "spawn_job"

        except LemmingsStop as stop:
            _handle_exception(lemmings.lemmings_job,lemmings.database,
                                stop, lemmings_stop= True)
        except Exception as any_other_exception:
            _handle_exception(lemmings.lemmings_job,lemmings.database,
                                any_other_exception, lemmings_stop= False)
    else:
        try:
            # check_on_end required if user does database updates in it
            # also check if user condition is reached and update database param
            condition_reached = lemmings.lemmings_job.check_on_end()
            #--testing------#
            # lemmings.database.update_previous_loop('condition_reached',
            #                                 condition_reached)
            #--end testing--#
            lemmings.database.update_current_loop('condition_reached',
                                                condition_reached)
            lemmings.lemmings_job.end_message = "Target CPU limit reached"
            lemmings.database.update_current_loop('end_message',
                            lemmings.lemmings_job.end_message)
            lemmings.lemmings_job.after_end_job()
            _check_and_activate_parallel_mode(lemmings.lemmings_job, "monitor")
            lemmings.write_log_file()


            lemmings.lemmings_job.status = "exit"
        except LemmingsStop as stop:
            _handle_exception(lemmings.lemmings_job,lemmings.database,
                                stop, lemmings_stop= True)
        except Exception as any_other_exception:
            _handle_exception(lemmings.lemmings_job,lemmings.database,
                                any_other_exception, lemmings_stop= False)

def _check_correct_parallel_settings(lemmings_job):
    """ Controlling whether everything correctly set by user for 'farming' mode of lemmings

        Input:
            :lemmings_job: lemmings job object

        Output:
            :num_workflow: int, number of separate workflows to consider
    """

    try:
        if not 'parallel_mode' in lemmings_job.machine.user.expert_params:
            raise KeyError
        if lemmings_job.machine.user.expert_params["parallel_mode"] is not True:
            raise ValueError
    except ValueError as excep:
        raise LemmingsStop("Parallel mode not activated, please do so through\n"
            + "expert_params:\n"
            + "  parallel_mode: True\n"
            + "\n"
            + "in the workflow.yml")
    except KeyError as excep:
        raise LemmingsStop("Parallel mode key not specified, please do so through\n"
            + "expert_params:\n"
            + "  parallel_mode: True\n"
            + "\n"
            + "in your workflow.yml")

    try:
        num_workflows = len(lemmings_job.machine.user.parallel_params['parameter_array'])
    except KeyError:
        raise LemmingsStop("parameter array not or wrongly specified, please use this structure:\n"
                        + "parallel_params:\n"
                        + "  parameter_array: \n"
                        + "  - par1: value \n"
                        + "    par2: value \n"
                        + "  - par1: value \n"
                        + "    par3: value \n"
                        + "  - par2: value \n"
                        + "    par4: value \n"
                        + "\n"
                        + "in your {workflow}.yml")

    return num_workflows

def _generate_workflow_replicates(workflow_nbs, lemmings_job, database):
    """ Function performing the actual replication of the current work folder structure
        for separate lemmings calls

        Input:
            :workflow_nbs: int, number of copies to create
            :lemmings_job: lemmings_job object
            :database:    Database object  

        Output:
            :workflows_list: list of type string containing workflow folders generated
    """

    workflows_list = []

    dir_info = os.listdir()
    for ii in np.arange(workflow_nbs):
        tmp_workflow = "WF_%03d" %ii
        # In case the user specified a name for his workflow
        try:
            tmp_workflow += ("_"+
                lemmings_job.machine.user.parallel_params['parameter_array'][ii]["workflow_name"])
        except KeyError as excep:   
            pass

        try:
            if os.path.isdir(tmp_workflow) and lemmings_job.machine.user.parallel_params["overwrite_dirs"]:
                shutil.rmtree(tmp_workflow)
            os.mkdir(tmp_workflow)
        except KeyError:
            raise LemmingsStop("Overwrite directory option not specified, please do so through\n"
                            + "parallel_params:\n"
                            + "  overwrite_dirs: True or False\n"
                            + "\n"
                            + "in the workflow.yml")
        except FileExistsError as excep:
            _handle_exception(lemmings_job,database,
                                    excep, lemmings_stop= False)
            sys.exit()

        for item in dir_info:
            if os.path.isfile(item):
                if item not in [database.db_path.split("./")[-1],
                                    lemmings_job.workflow+".yml"]:
                    shutil.copy(item, tmp_workflow)
            else:
                if item not in [database.latest_chain_name]:
                    if item.split('_')[0] != "Workflow" and item not in database.get_chain_names():
                        lemmings_job.pathtools.copy_dir(item, tmp_workflow+"/")

        workflows_list.append(tmp_workflow)
        
    return workflows_list

def _launch_workflows_parallel(workflows_list, lemmings_job, database):
    """ Function performing the actual replication of the current work folder structure
        for separate lemmings calls

        Input:
            :workflow_list: list of type string containing names of workflows to consider
            :lemmings_job: lemmings_job object
            :database:    Database object  

        Output:
            :max_par_wf: None or int with number of max nb of simultaneous workflows to launch
    """

    tmp_path = os.getcwd()
    for ii, workflow in enumerate(workflows_list):
        # go to subdirectory to launch lemmings
        os.chdir(workflow)

        max_par_wf = None
        try:
            max_par_wf = lemmings_job.machine.user.parallel_params["max_parallel_workflows"]
            if not ii+1 > max_par_wf:
                subprocess.call("lemmings-hpc run " + lemmings_job.workflow + " --noverif"
                                + " --yaml=../"+lemmings_job.workflow+".yml" , shell = True)
            # go back to main directory
            os.chdir(tmp_path)
            if not ii+1 > max_par_wf:
                database.update_nested_level_to_loop('parallel_runs',workflow, 'start')
            else:
                # case in which we put them on hold for submission at later stage
                database.update_nested_level_to_loop('parallel_runs',workflow, 'wait')
        except KeyError:
            subprocess.call("lemmings-hpc run " + lemmings_job.workflow + " --noverif"
                            + " --yaml=../"+lemmings_job.workflow+".yml" , shell = True)
            os.chdir(tmp_path)
            database.update_nested_level_to_loop('parallel_runs',workflow, 'start')
        except TypeError:
            subprocess.call("lemmings-hpc run " + lemmings_job.workflow + " --noverif"
                            + " --yaml=../"+lemmings_job.workflow+".yml" , shell = True)
            os.chdir(tmp_path)
            database.update_nested_level_to_loop('parallel_runs',workflow, 'start')

    return max_par_wf

def _handle_exception(lemmings_job, database, exception, lemmings_stop = False):
    """ Function that performs the updates in case an exception is raised through
        LemmingsStop or other

        Input:
            lemmings_job: lemmings_job class object
            database: database class object
            exception: raised exception class message
            lemmings_stop: boolean, whether exception raised through LemmingsStop or not
        Output:
            None: performs updates in database and ends the lemmings chain
    """

    end_msg = str(exception)
    if not lemmings_stop:
        end_msg = "Unexpected exception: " + end_msg + "\n" + traceback.format_exc()
        traceback.print_exc()

    lemmings_job.end_message = end_msg
    database.update_current_loop("end_message", lemmings_job.end_message)
    lemmings_job.status = "exit"


def _check_and_activate_parallel_mode(lemmings_job, stage):
    """ Function that checks if the parallel mode has been activated and acts
        accordingly

        Input:
            lemmings_job: lemmings_job class object
            stage: str, which parallel functionality to call
                    = create: create_replicate_workflows()
                    = monitor: monitor_replicate_workflows()

        Output:
            None: launches parallel mode functions if parallel mode activated
    """

    if hasattr(lemmings_job.machine.user, 'expert_params'):
        if 'parallel_mode' in lemmings_job.machine.user.expert_params:
            if lemmings_job.machine.user.expert_params['parallel_mode']:
                if stage == "create": # moved call to cli.py
                    lemmings_job.create_replicate_workflows()
                elif stage == "monitor":
                    lemmings_job.monitor_replicate_workflows()
                else:
                    raise LemmingsStop("Unknown parallel stage function")


    # def write_log_file(self):
    #     """write the log file"""
    #     chain_name = self.database.latest_chain_name
    #     database = self.database._database

    #     usefull_keys = ['datetime', 'job_id', 'pjob_id', 'init_path', 'temporal_path', 'dtsum']
    #     columns_len = [20, 30]

    #     if chain_name is None:
    #         print("No chain found. Check database file in your current directory...")
    #     else:
    #         log_msg = ("\n#########################################\n"
    #                    + "         # Lemmings Version : "
    #                    + str(database[chain_name][0]['lemmings_version']) + " #         \n"
    #                    + "#########################################\n\n")


    #         for i, loop in enumerate(database[chain_name]):
    #             log_msg += "\n\n"
    #             log_msg += self.write_head(columns_len, i+1)

    #             for key in loop:
    #                 if key in usefull_keys:
    #                     key_blank_nb = self.adjust_column_size(key, size=20)
    #                     val_blank_nb = self.adjust_column_size(loop[key], size=30)
    #                     log_msg += ('|' + str(key) + key_blank_nb * " "
    #                                 + str(loop[key]) + val_blank_nb * " " + '|' + '\n')

    #                     log_msg += "├" + (sum(columns_len)) * " " + "┤"  + '\n'

    #             log_msg += "├" + (sum(columns_len)) * "-" + "┤"

    #         log_msg += "\n\n"
    #         log_msg += self.write_whole_chain_param(database, chain_name)

    #     with open(os.path.join(chain_name, chain_name + '.log'), 'w') as fout:
    #         fout.write(log_msg)


    # def adjust_column_size(self, key, size):

    #     word_len = len(str(key))
    #     return(size - word_len)

    # def write_head(self, columns_len, loop_nb):
    #     tot_size = sum(columns_len)
    #     log_msg =  "\n├" + (tot_size) * "-" + "┤" +'\n'

    #     log_msg += ('|' + (int(tot_size/2)-4)* " " + "LOOP N° " + str(loop_nb) + '|'
    #                 + self.adjust_column_size("LOOP N°", int(tot_size/2)) * " " + '\n')
    #     log_msg += "├" + tot_size * '-' + "┤" +'\n'
    #     return log_msg

    # def write_whole_chain_param(self, database, chain_name):

    #     log_msg = ("\n\nTotal physical time [s]   " + str(database[chain_name][-2]['dtsum']) + '\n'
    #                + "Total CPU cost [hours]    " + str(database[chain_name][-2]['end_cpu_time']) + '\n')


    #     if database[chain_name][-1]['safe_stop'] == True:
    #         log_msg += "Lemmings STOP because using 'safe stop' command "
    #     if 'run_crash' in database[chain_name][-2]:
    #         log_msg += "Your run CRASH, see avbp.o file"
    #     if 'condition_reached' in database[chain_name][-2]:
    #         log_msg += ("Condition " + str(database[chain_name][0]['simu_end_time'])
    #                     + " [s] is    REACHED")

    #     return log_msg
